import pytest

from dbt.tests.adapter.ephemeral.test_ephemeral import (
    BaseEphemeralErrorHandling,
    BaseEphemeralMulti,
    BaseEphemeralNested,
)


@pytest.mark.skip(reason="Nested CTEs not supported in Views")
class TestEphemeral(BaseEphemeralMulti):
    pass


@pytest.mark.skip(reason="Nested CTEs not supported in Views")
class TestEphemeralNested(BaseEphemeralNested):
    pass


class TestEphemeralErrorHandling(BaseEphemeralErrorHandling):
    pass



import pytest

from dbt.tests.adapter.aliases.fixtures import MACROS__EXPECT_VALUE_SQL
from dbt.tests.adapter.aliases.test_aliases import (
    BaseAliasErrors,
    BaseAliases,
    BaseSameAliasDifferentDatabases,
    BaseSameAliasDifferentSchemas,
)


class TestAliasesFabric(BaseAliases):
    @pytest.fixture(scope="class")
    def macros(self):
        return {"expect_value.sql": MACROS__EXPECT_VALUE_SQL}


class TestAliasErrorsFabric(BaseAliasErrors):
    @pytest.fixture(scope="class")
    def macros(self):
        return {"expect_value.sql": MACROS__EXPECT_VALUE_SQL}


class TestSameAliasDifferentSchemasFabric(BaseSameAliasDifferentSchemas):
    @pytest.fixture(scope="class")
    def macros(self):
        return {"expect_value.sql": MACROS__EXPECT_VALUE_SQL}


class TestSameAliasDifferentDatabasesFabric(BaseSameAliasDifferentDatabases):
    @pytest.fixture(scope="class")
    def macros(self):
        return {"expect_value.sql": MACROS__EXPECT_VALUE_SQL}



import datetime as dt
import struct
import time
from contextlib import contextmanager
from typing import Any

import agate
import dbt_common.exceptions
import pyodbc
from dbt_common.clients.agate_helper import empty_table
from dbt_common.events.contextvars import get_node_info
from dbt_common.events.functions import fire_event
from dbt_common.utils.casting import cast_to_str

from dbt.adapters.contracts.connection import AdapterResponse, Connection, ConnectionState
from dbt.adapters.events.logging import AdapterLogger
from dbt.adapters.events.types import ConnectionUsed, SQLQuery, SQLQueryStatus
from dbt.adapters.fabric import __version__
from dbt.adapters.fabric.fabric_api_client import FabricApiClient
from dbt.adapters.fabric.fabric_credentials import FabricCredentials
from dbt.adapters.fabric.fabric_token_provider import FabricTokenProvider
from dbt.adapters.sql import SQLConnectionManager

logger = AdapterLogger("fabric")

# https://github.com/mkleehammer/pyodbc/wiki/Data-Types
datatypes = {
    "str": "varchar",
    "uuid.UUID": "uniqueidentifier",
    "uuid": "uniqueidentifier",
    "float": "bigint",
    "int": "int",
    "bytes": "varbinary",
    "bytearray": "varbinary",
    "bool": "bit",
    "datetime.date": "date",
    "datetime.datetime": "datetime2(6)",
    "datetime.time": "time",
    "decimal.Decimal": "decimal",
}


def bool_to_connection_string_arg(key: str, value: bool) -> str:
    """
    Convert a boolean to a connection string argument.

    Parameters
    ----------
    key : str
        The key to use in the connection string.
    value : bool
        The boolean to convert.

    Returns
    -------
    out : str
        The connection string argument.
    """
    return f"{key}={'Yes' if value else 'No'}"


def byte_array_to_datetime(value: bytes) -> dt.datetime:
    """
    Converts a DATETIMEOFFSET byte array to a timezone-aware datetime object

    Parameters
    ----------
    value : buffer
        A binary value conforming to SQL_SS_TIMESTAMPOFFSET_STRUCT

    Returns
    -------
    out : datetime

    Source
    ------
    SQL_SS_TIMESTAMPOFFSET datatype and SQL_SS_TIMESTAMPOFFSET_STRUCT layout:
    https://learn.microsoft.com/sql/relational-databases/native-client-odbc-date-time/data-type-support-for-odbc-date-and-time-improvements
    """
    # unpack 20 bytes of data into a tuple of 9 values
    tup = struct.unpack("<6hI2h", value)

    # construct a datetime object
    return dt.datetime(
        year=tup[0],
        month=tup[1],
        day=tup[2],
        hour=tup[3],
        minute=tup[4],
        second=tup[5],
        microsecond=tup[6] // 1000,  # https://bugs.python.org/issue15443
        tzinfo=dt.timezone(dt.timedelta(hours=tup[7], minutes=tup[8])),
    )


class FabricConnectionManager(SQLConnectionManager):
    TYPE = "fabric"
    _host: str | None = None
    _fabric_token_provider: FabricTokenProvider | None = None
    _fabric_api_client: FabricApiClient | None = None

    @classmethod
    def get_fabric_token_provider(cls, credentials: FabricCredentials) -> FabricTokenProvider:
        if cls._fabric_token_provider is None:
            cls._fabric_token_provider = FabricTokenProvider(credentials)
        return cls._fabric_token_provider

    @classmethod
    def get_fabric_api_client(cls, credentials: FabricCredentials) -> FabricApiClient:
        if cls._fabric_api_client is None:
            cls._fabric_api_client = FabricApiClient(
                credentials, cls.get_fabric_token_provider(credentials)
            )
        return cls._fabric_api_client

    @contextmanager
    def exception_handler(self, sql):
        try:
            yield

        except pyodbc.DatabaseError as e:
            logger.debug("Database error: {}".format(str(e)))

            try:
                # attempt to release the connection
                self.release()
            except pyodbc.Error:
                logger.debug("Failed to release connection!")

            raise dbt_common.exceptions.DbtDatabaseError(str(e).strip()) from e

        except Exception as e:
            logger.debug(f"Error running SQL: {sql}")
            logger.debug("Rolling back transaction.")
            self.release()
            if isinstance(e, dbt_common.exceptions.DbtRuntimeError):
                # during a sql query, an internal to dbt exception was raised.
                # this sounds a lot like a signal handler and probably has
                # useful information, so raise it without modification.
                raise

            raise dbt_common.exceptions.DbtRuntimeError(e)

    @classmethod
    def get_host(cls, credentials: FabricCredentials) -> str:
        if cls._host is None:
            if credentials.host:
                cls._host = credentials.host
            elif credentials.workspace_id or credentials.workspace_name:
                api = cls.get_fabric_api_client(credentials)
                cls._host = api.get_warehouse_connection_string()
            else:
                raise dbt_common.exceptions.DbtConfigError(
                    "Either host or workspace_id must be provided."
                )
        assert cls._host is not None
        return cls._host

    @classmethod
    def open(cls, connection: Connection) -> Connection:
        if connection.state == ConnectionState.OPEN:
            logger.debug("Connection is already open, skipping open.")
            return connection

        credentials = cls.get_credentials(connection.credentials)

        con_str = [f"DRIVER={{{credentials.driver}}}"]

        con_str.append(f"SERVER={cls.get_host(credentials)}")

        con_str.append(f"Database={credentials.database}")
        con_str.append("Pooling=true")

        # Enabling trace flag
        if credentials.trace_flag:
            con_str.append("SQL_ATTR_TRACE=SQL_OPT_TRACE_ON")
        else:
            con_str.append("SQL_ATTR_TRACE=SQL_OPT_TRACE_OFF")

        assert credentials.authentication is not None

        if "ActiveDirectory" in credentials.authentication:
            con_str.append(f"Authentication={credentials.authentication}")

            if credentials.authentication == "ActiveDirectoryPassword":
                con_str.append(f"UID={{{credentials.UID}}}")
                con_str.append(f"PWD={{{credentials.PWD}}}")
            if credentials.authentication == "ActiveDirectoryServicePrincipal":
                con_str.append(f"UID={{{credentials.client_id}}}")
                con_str.append(f"PWD={{{credentials.client_secret}}}")
            elif credentials.authentication == "ActiveDirectoryInteractive":
                con_str.append(f"UID={{{credentials.UID}}}")

        elif credentials.windows_login:
            con_str.append("trusted_connection=Yes")
        elif credentials.authentication == "sql":
            raise pyodbc.DatabaseError("SQL Authentication is not supported by Microsoft Fabric")

        # https://docs.microsoft.com/en-us/sql/relational-databases/native-client/features/using-encryption-without-validation?view=sql-server-ver15
        assert credentials.encrypt is not None
        assert credentials.trust_cert is not None

        con_str.append(bool_to_connection_string_arg("encrypt", credentials.encrypt))
        con_str.append(
            bool_to_connection_string_arg("TrustServerCertificate", credentials.trust_cert)
        )

        plugin_version = __version__.version
        application_name = f"dbt-{credentials.type}/{plugin_version}"
        con_str.append(f"APP={application_name}")

        try:
            con_str.append(f"ConnectRetryCount={credentials.retries}")
            con_str.append("ConnectRetryInterval=10")

        except Exception as e:
            logger.debug(
                "Retry count should be a integer value. Skipping retries in the connection string.",
                str(e),
            )

        con_str_concat = ";".join(con_str)

        index = []
        for i, elem in enumerate(con_str):
            if "pwd=" in elem.lower():
                index.append(i)

        if len(index) != 0:
            con_str[index[0]] = "PWD=***"

        con_str_display = ";".join(con_str)

        retryable_exceptions = [  # https://github.com/mkleehammer/pyodbc/wiki/Exceptions
            pyodbc.InternalError,  # not used according to docs, but defined in PEP-249
            pyodbc.OperationalError,
            pyodbc.InterfaceError,
        ]

        def connect():
            logger.debug(f"Using connection string: {con_str_display}")
            pyodbc.pooling = True
            pyodbc.odbcversion = "3.8"
            attrs_before = cls.get_fabric_token_provider(credentials).get_pyodbc_attributes()

            handle = pyodbc.connect(
                con_str_concat,
                attrs_before=attrs_before,
                autocommit=True,
                timeout=credentials.login_timeout,
            )
            handle.timeout = credentials.query_timeout
            logger.debug(f"Connected to db: {credentials.database}")
            return handle

        return cls.retry_connection(
            connection,
            connect=connect,
            logger=logger,
            retry_limit=credentials.retries,
            retryable_exceptions=retryable_exceptions,
        )

    def cancel(self, connection: Connection):
        pass

    def add_begin_query(self) -> tuple[Connection, Any]:
        pass

    def add_commit_query(self) -> tuple[Connection, Any]:
        pass

    def add_query(
        self,
        sql: str,
        auto_begin: bool = True,
        bindings: Any | None = None,
        abridge_sql_log: bool = False,
    ) -> tuple[Connection, Any]:
        connection = self.get_thread_connection()

        if auto_begin and connection.transaction_open is False:
            self.begin()

        fire_event(
            ConnectionUsed(
                conn_type=self.TYPE,
                conn_name=cast_to_str(connection.name),
                node_info=get_node_info(),
            )
        )

        with self.exception_handler(sql):
            if abridge_sql_log:
                log_sql = "{}...".format(sql[:512])
            else:
                log_sql = sql

            fire_event(
                SQLQuery(
                    conn_name=cast_to_str(connection.name), sql=log_sql, node_info=get_node_info()
                )
            )

            pre = time.time()

            cursor = connection.handle.cursor()

            # pyodbc does not handle a None type binding!
            if bindings is None:
                cursor.execute(sql)
            else:
                bindings = [
                    binding if not isinstance(binding, dt.datetime) else binding.isoformat()
                    for binding in bindings
                ]
                cursor.execute(sql, bindings)

            # convert DATETIMEOFFSET binary structures to datetime ojbects
            # https://github.com/mkleehammer/pyodbc/issues/134#issuecomment-281739794
            connection.handle.add_output_converter(-155, byte_array_to_datetime)

            fire_event(
                SQLQueryStatus(
                    status=str(self.get_response(cursor)),
                    elapsed=round((time.time() - pre)),
                    node_info=get_node_info(),
                )
            )

            return connection, cursor

    @classmethod
    def get_credentials(cls, credentials: FabricCredentials) -> FabricCredentials:
        return credentials

    @classmethod
    def get_response(cls, cursor: pyodbc.Cursor) -> AdapterResponse:
        message = "\n".join(msg[1] for msg in cursor.messages) if cursor.messages else ""
        return AdapterResponse(
            _message=message,
            rows_affected=cursor.rowcount,
        )

    @classmethod
    def data_type_code_to_name(cls, type_code: int | str) -> str:
        data_type = str(type_code)[str(type_code).index("'") + 1 : str(type_code).rindex("'")]
        return datatypes[data_type]

    def execute(
        self, sql: str, auto_begin: bool = True, fetch: bool = False, limit: int | None = None
    ) -> tuple[AdapterResponse, agate.Table]:
        sql = self._add_query_comment(sql)
        _, cursor = self.add_query(sql, auto_begin)
        response = self.get_response(cursor)
        if fetch:
            # Get the result of the first non-empty result set (if any)
            while cursor.description is None:
                if not cursor.nextset():
                    break
            table = self.get_result_from_cursor(cursor, limit)
        else:
            table = empty_table()
        # Step through all result sets so we process all errors
        while cursor.nextset():
            pass
        return response, table



import pytest

from dbt.tests.adapter.utils.test_dateadd import BaseDateAdd


class TestDateAddFabric(BaseDateAdd):
    @pytest.fixture(scope="class")
    def project_config_update(self):
        return {}



import os
from pathlib import Path

import pytest
import yaml

from dbt.adapters.fabric.fabric_api_client import FabricApiClient
from dbt.adapters.fabric.fabric_credentials import FabricCredentials
from dbt.adapters.fabric.fabric_token_provider import FabricTokenProvider

pytest_plugins = ["dbt.tests.fixtures.project"]


@pytest.fixture(scope="class")
def dbt_profile_target(dbt_profile_target_update):
    target = {
        "type": "fabric",
        "driver": os.getenv("FABRIC_TEST_DRIVER", "ODBC Driver 18 for SQL Server"),
        "host": os.getenv("FABRIC_TEST_HOST"),
        "workspace_name": os.getenv("FABRIC_TEST_WORKSPACE_NAME"),
        "workspace_id": os.getenv("FABRIC_TEST_WORKSPACE_ID"),
        "lakehouse_id": os.getenv("FABRIC_TEST_LAKEHOUSE_ID"),
        "lakehouse_name": os.getenv("FABRIC_TEST_LAKEHOUSE_NAME"),
        "authentication": "auto",
        "database": os.getenv("FABRIC_TEST_DWH_NAME"),
        "retries": 3,
        "threads": int(os.getenv("FABRIC_TEST_THREADS", 20)),
        "login_timeout": 30,
        "query_timeout": 10,
    }

    target.update(dbt_profile_target_update)
    return target


@pytest.fixture(scope="class")
def dbt_profile_target_update():
    return {}


@pytest.fixture(scope="class")
def profile_user(dbt_profile_target):
    return "dbo"


def pytest_addoption(parser):
    parser.addoption("--with-grants", action="store_true", default=False, help="run GRANT tests")


def pytest_configure(config):
    config.addinivalue_line("markers", "grants: mark test containing GRANT statements")


def pytest_collection_modifyitems(config, items):
    if config.getoption("--with-grants"):
        # --with-grants given in cli: do not skip slow tests
        return
    skip_grants = pytest.mark.skip(reason="need --with-grants option to run")
    for item in items:
        if "grants" in item.keywords:
            item.add_marker(skip_grants)


@pytest.fixture(scope="class")
def logs_dir(request, prefix):
    dbt_log_dir = os.path.join(request.config.rootdir, "logs", prefix)
    print(f"\n=== Test logs_dir: {dbt_log_dir}\n")
    os.environ["DBT_LOG_PATH"] = str(dbt_log_dir)
    yield str(Path(dbt_log_dir))
    del os.environ["DBT_LOG_PATH"]


@pytest.fixture(scope="class")
def dbt_core_bug_workaround(project):
    # Workaround for https://github.com/dbt-labs/dbt-core/issues/5410
    with open(Path(project.project_root).parent / "dbt_project.yml", "w") as f:
        f.write(yaml.safe_dump({"name": "workaround"}))


@pytest.fixture(scope="class")
def project(
    project_setup,
    project_files,
):
    from dbt.tests.fixtures.project import TestProjInfo

    class TestProjInfoFabric(TestProjInfo):
        def get_tables_in_schema(self):
            sql = f"""
                    select
                            t.name as table_name,
                            'table' as materialization
                    from sys.tables t
                    inner join sys.schemas s
                    on s.schema_id = t.schema_id
                    where lower(s.name) = '{self.test_schema.lower()}'
                    union all
                    select
                            v.name as table_name,
                            'view' as materialization
                    from sys.views v
                    inner join sys.schemas s
                    on s.schema_id = v.schema_id
                    where lower(s.name) = '{self.test_schema.lower()}'
                    """
            result = self.run_sql(sql, fetch="all")
            return {model_name: materialization for (model_name, materialization) in result}

    return TestProjInfoFabric(
        project_root=project_setup.project_root,
        profiles_dir=project_setup.profiles_dir,
        adapter_type=project_setup.adapter_type,
        test_dir=project_setup.test_dir,
        shared_data_dir=project_setup.shared_data_dir,
        test_data_dir=project_setup.test_data_dir,
        test_schema=project_setup.test_schema,
        database=project_setup.database,
        test_config=project_setup.test_config,
    )


@pytest.fixture(scope="class")
def credentials(adapter) -> FabricCredentials:
    return adapter.config.credentials


@pytest.fixture(scope="class")
def fabric_token_provider(credentials: FabricCredentials) -> FabricTokenProvider:
    return FabricTokenProvider(credentials)


@pytest.fixture(scope="class")
def fabric_api_client(
    fabric_token_provider: FabricTokenProvider, credentials: FabricCredentials
) -> FabricApiClient:
    return FabricApiClient.create(credentials, fabric_token_provider)



import pytest

from dbt.tests.util import run_dbt


class BaseDbtPackageTests:
    @pytest.fixture(scope="class")
    def package_name(self) -> str:
        raise NotImplementedError("Subclasses must implement package_name")

    @pytest.fixture(scope="class")
    def package_repo(self) -> str:
        raise NotImplementedError("Subclasses must implement package_repo")

    @pytest.fixture(scope="class")
    def package_revision(self) -> str:
        raise NotImplementedError("Subclasses must implement package_version")

    @pytest.fixture(scope="class")
    def packages(self, package_name: str, package_repo: str, package_revision: str):
        packages = [
            {"git": package_repo, "revision": package_revision},
            {
                "git": package_repo,
                "revision": package_revision,
                "subdirectory": "integration_tests",
            },
        ]

        if package_name != "dbt_utils":
            packages.append({"package": "dbt-labs/dbt_utils", "version": "1.3.0"})

        return {"packages": packages}

    @pytest.fixture(scope="class")
    def models_config(self):
        return {}

    @pytest.fixture(scope="class")
    def seeds_config(self):
        return {}

    @pytest.fixture(scope="class")
    def tests_config(self):
        return {}

    @pytest.fixture(scope="class")
    def project_config_update(
        self, package_name: str, models_config: dict, seeds_config: dict, tests_config: dict
    ):
        dispatches = [
            {
                "macro_namespace": package_name,
                "search_order": [
                    f"test_dbt_package",
                    "dbt",
                    package_name,
                ],
            }
        ]
        if package_name != "dbt_utils":
            dispatches.append(
                {
                    "macro_namespace": "dbt_utils",
                    "search_order": [
                        f"test_dbt_package",
                        "dbt",
                        "dbt_utils",
                    ],
                }
            )

        return {
            "name": "test_dbt_package",
            "vars": {"dbt_date:time_zone": "UTC"},
            "dispatch": dispatches,
            "seeds": seeds_config,
            "models": models_config,
            "tests": tests_config,
        }

    def test_package(self, project, dbt_core_bug_workaround):
        run_dbt(["deps"])
        run_dbt(["seed"])
        run_dbt(["run"])



import pytest

from dbt.tests.adapter.empty._models import schema_sources_yml
from dbt.tests.adapter.empty.test_empty import BaseTestEmpty, BaseTestEmptyInlineSourceRef


class TestFabricEmpty(BaseTestEmpty):
    pass


class TestFabricEmptyInlineSourceRef(BaseTestEmptyInlineSourceRef):
    model_inline_sql = """
        select * from {{ source('seed_sources', 'raw_source') }} as raw_source
        """

    @pytest.fixture(scope="class")
    def models(self):
        return {
            "model.sql": "select * from {{ source('seed_sources', 'raw_source') }}",
            "sources.yml": schema_sources_yml,
        }



import struct
import time
from itertools import chain, repeat

from azure.core.credentials import AccessToken
from azure.identity import (
    AzureCliCredential,
    ClientSecretCredential,
    DefaultAzureCredential,
    EnvironmentCredential,
)

from dbt.adapters.fabric.fabric_credentials import FabricCredentials


def get_mssparkutils_access_token(scope: str) -> AccessToken:
    from notebookutils import credentials

    aad_token = credentials.getToken(scope)
    expires_on = int(time.time() + 4500.0)
    token = AccessToken(
        token=aad_token,
        expires_on=expires_on,
    )
    return token


def get_cli_access_token(scope: str) -> AccessToken:
    token = AzureCliCredential().get_token(scope)
    return token


def get_auto_access_token(scope: str) -> AccessToken:
    token = DefaultAzureCredential().get_token(scope)
    return token


def get_environment_access_token(scope: str) -> AccessToken:
    token = EnvironmentCredential().get_token(scope)
    return token


azure_auth_functions = {
    "cli": get_cli_access_token,
    "auto": get_auto_access_token,
    "environment": get_environment_access_token,
    "synapsespark": get_mssparkutils_access_token,
    "fabricspark": get_mssparkutils_access_token,
}


class FabricTokenProvider:
    AZURE_CREDENTIAL_SCOPE = "https://database.windows.net/.default"
    FABRIC_CREDENTIAL_SCOPE = "https://analysis.windows.net/powerbi/api/.default"
    SYNAPSE_SPARK_CREDENTIAL_SCOPE = "DW"
    FABRIC_SPARK_CREDENTIAL_SCOPE = "pbi"
    _tokens: dict[str, AccessToken] = {}
    SQL_COPT_SS_ACCESS_TOKEN = 1256

    def __init__(self, credentials: FabricCredentials):
        self.credentials = credentials

    def get_token_scope(self) -> str:
        if self.credentials.token_scope:
            return self.credentials.token_scope

        if not self.credentials.host and (
            self.credentials.workspace_id or self.credentials.workspace_name
        ):
            return self.FABRIC_CREDENTIAL_SCOPE
        if "synapse" in self.credentials.authentication.lower():
            return self.SYNAPSE_SPARK_CREDENTIAL_SCOPE
        if "fabric" in self.credentials.authentication.lower():
            return self.FABRIC_SPARK_CREDENTIAL_SCOPE
        if "azuresynapse.net" in self.credentials.host.lower():
            return self.SYNAPSE_SPARK_CREDENTIAL_SCOPE
        if "fabric.microsoft.com" in self.credentials.host.lower():
            return self.FABRIC_CREDENTIAL_SCOPE
        if "database.windows.net" in self.credentials.host.lower():
            return self.AZURE_CREDENTIAL_SCOPE
        return self.FABRIC_CREDENTIAL_SCOPE

    def get_api_token(self) -> str | None:
        return self._get_token(usage_is_sql=False)

    def get_sql_token(self, scope: str | None = None) -> str | None:
        return self._get_token(scope=scope, usage_is_sql=True)

    def _get_token(self, scope: str | None = None, usage_is_sql: bool = False) -> str | None:
        MAX_REMAINING_TIME = 300

        if self.credentials.access_token:
            return self.credentials.access_token

        _scope = self.FABRIC_CREDENTIAL_SCOPE if not usage_is_sql else None
        scope = scope or self.credentials.token_scope or _scope or self.get_token_scope()
        authentication = str(self.credentials.authentication).lower()

        current_token = self._tokens.get(scope)
        time_remaining = (
            (current_token.expires_on - time.time()) if current_token else MAX_REMAINING_TIME
        )

        if usage_is_sql and authentication not in azure_auth_functions:
            return None
        if current_token and time_remaining >= MAX_REMAINING_TIME:
            return current_token.token
        if authentication in azure_auth_functions:
            azure_auth_function = azure_auth_functions[authentication]
            token = azure_auth_function(scope)
            self._tokens[scope] = token
            return token.token
        if authentication == "activedirectoryserviceprincipal" and not usage_is_sql:
            client_id = self.credentials.client_id
            client_secret = self.credentials.client_secret
            tenant_id = self.credentials.tenant_id
            cred = ClientSecretCredential(
                client_id=client_id,
                client_secret=client_secret,
                tenant_id=tenant_id,
            )
            token = cred.get_token(scope)
            self._tokens[scope] = token
            return token.token

        return None

    @staticmethod
    def convert_bytes_to_mswindows_byte_string(value: bytes) -> bytes:
        encoded_bytes = bytes(chain.from_iterable(zip(value, repeat(0))))
        return struct.pack("<i", len(encoded_bytes)) + encoded_bytes

    @staticmethod
    def convert_access_token_to_mswindows_byte_string(token: str) -> bytes:
        value = bytes(token, "UTF-8")
        return FabricTokenProvider.convert_bytes_to_mswindows_byte_string(value)

    def get_pyodbc_attributes(self) -> dict[int, bytes]:
        attrs_before: dict
        token = self.get_sql_token()
        if token:
            token_bytes = self.convert_access_token_to_mswindows_byte_string(token)
            attrs_before = {self.SQL_COPT_SS_ACCESS_TOKEN: token_bytes}
        else:
            attrs_before = {}
        return attrs_before



import pytest

from dbt.tests.adapter.unit_testing.test_case_insensitivity import BaseUnitTestCaseInsensivity
from dbt.tests.adapter.unit_testing.test_invalid_input import BaseUnitTestInvalidInput
from dbt.tests.adapter.unit_testing.test_types import BaseUnitTestingTypes


class TestFabricUnitTestingTypes(BaseUnitTestingTypes):
    @pytest.fixture
    def data_types(self):
        # sql_value, yaml_value
        return [
            ["1", "1"],
            ["'1'", "1"],
            ["CAST('2020-01-02' AS DATE)", "2020-01-02"],
            ["CAST('2013-11-03 00:00:00-0' AS DATETIME2(6))", "2013-11-03 00:00:00-0"],
            ["CAST('1' AS INT)", "1"],
        ]


class TestFabricUnitTestCaseInsensivity(BaseUnitTestCaseInsensivity):
    pass


class TestFabricUnitTestInvalidInput(BaseUnitTestInvalidInput):
    pass



import json

import pytest

from dbt.tests.util import run_dbt, run_dbt_and_capture

NUM_VIEWS = 10
NUM_EXPECTED_RELATIONS = 1 + NUM_VIEWS

TABLE_BASE_SQL = """
{{ config(materialized='table') }}

select 1 as id
""".lstrip()

VIEW_X_SQL = """
select id from {{ ref('my_model_base') }}
""".lstrip()

# TODO - fix the call {% set relation_list_result = fabric__list_relations_without_caching(schema_relation) %}
MACROS__VALIDATE__FABRIC__LIST_RELATIONS_WITHOUT_CACHING = """
{% macro validate_list_relations_without_caching(schema_relation) -%}

    {% call statement('list_relations_without_caching', fetch_result=True) -%}
        with base as (
        select
            DB_NAME() as [database],
            t.name as [name],
            SCHEMA_NAME(t.schema_id) as [schema],
            'table' as table_type
        from sys.tables as t {{ information_schema_hints() }}
        union all
        select
            DB_NAME() as [database],
            v.name as [name],
            SCHEMA_NAME(v.schema_id) as [schema],
            'view' as table_type
        from sys.views as v {{ information_schema_hints() }}
        )
        select * from base
        where [schema] like '{{ schema_relation }}'
    {% endcall %}

    {% set relation_list_result = load_result('list_relations_without_caching').table %}
    {% set n_relations = relation_list_result | length %}
    {{ log("n_relations: " ~ n_relations) }}
{% endmacro %}
"""


def parse_json_logs(json_log_output):
    parsed_logs = []
    for line in json_log_output.split("\n"):
        try:
            log = json.loads(line)
        except ValueError:
            continue

        parsed_logs.append(log)

    return parsed_logs


def find_result_in_parsed_logs(parsed_logs, result_name):
    return next(
        (
            item["data"]["msg"]
            for item in parsed_logs
            if result_name in item["data"].get("msg", "msg")
        ),
        False,
    )


def find_exc_info_in_parsed_logs(parsed_logs, exc_info_name):
    return next(
        (
            item["data"]["exc_info"]
            for item in parsed_logs
            if exc_info_name in item["data"].get("exc_info", "exc_info")
        ),
        False,
    )


class TestListRelationsWithoutCachingSingle:
    @pytest.fixture(scope="class")
    def models(self):
        my_models = {"my_model_base.sql": TABLE_BASE_SQL}
        for view in range(0, NUM_VIEWS):
            my_models.update({f"my_model_{view}.sql": VIEW_X_SQL})

        return my_models

    @pytest.fixture(scope="class")
    def macros(self):
        return {
            "validate_list_relations_without_caching.sql": MACROS__VALIDATE__FABRIC__LIST_RELATIONS_WITHOUT_CACHING,
        }

    def test__fabric__list_relations_without_caching(self, project):
        """
        validates that fabric__list_relations_without_caching
        macro returns a single record
        """
        run_dbt(["run", "-s", "my_model_base"])

        schemas = project.created_schemas

        for schema in schemas:
            kwargs = {"schema_relation": schema}
            _, log_output = run_dbt_and_capture(
                [
                    "--debug",
                    "run-operation",
                    "validate_list_relations_without_caching",
                    "--args",
                    str(kwargs),
                ]
            )
            assert "n_relations: 1" in log_output


class TestListRelationsWithoutCachingFull:
    @pytest.fixture(scope="class")
    def models(self):
        my_models = {"my_model_base.sql": TABLE_BASE_SQL}
        for view in range(0, NUM_VIEWS):
            my_models.update({f"my_model_{view}.sql": VIEW_X_SQL})

        return my_models

    @pytest.fixture(scope="class")
    def macros(self):
        return {
            "validate_list_relations_without_caching.sql": MACROS__VALIDATE__FABRIC__LIST_RELATIONS_WITHOUT_CACHING,
        }

    def test__fabric__list_relations_without_caching(self, project):
        # purpose of the first run is to create the replicated views in the target schema
        run_dbt(["run"])

        schemas = project.created_schemas

        for schema in schemas:
            kwargs = {"schema_relation": schema}
            _, log_output = run_dbt_and_capture(
                [
                    "--debug",
                    "run-operation",
                    "validate_list_relations_without_caching",
                    "--args",
                    str(kwargs),
                ]
            )
            assert f"n_relations: {NUM_EXPECTED_RELATIONS}" in log_output



import pytest

from dbt.tests.adapter.utils.fixture_date import models__test_date_yml
from dbt.tests.adapter.utils.test_date import BaseDate


class TestDateFabric(BaseDate):
    @pytest.fixture(scope="class")
    def models(self):
        models__test_date_sql = """
with generated_dates as (

    {{
        dbt.date_spine(
            "day",
            "cast('2023-09-07' as date)",
            "cast('2023-09-10' as date)"
        )
    }}

),

expected_dates as (

    select cast('2023-09-07' as date) as expected
    union all
    select cast('2023-09-08' as date) as expected
    union all
    select cast('2023-09-09' as date) as expected

),

joined as (
    select
        generated_dates.date_day,
        expected_dates.expected
    from generated_dates
    full outer join expected_dates on generated_dates.date_day = expected_dates.expected
)

select * from joined
"""

        return {
            "test_date.yml": models__test_date_yml,
            "test_date.sql": self.interpolate_macro_namespace(models__test_date_sql, "date"),
        }



import pytest

from dbt.tests.adapter.dbt_clone.test_dbt_clone import (
    BaseCloneNotPossible,
    BaseCloneSameTargetAndState,
)


class TestFabricCloneNotPossible(BaseCloneNotPossible):
    @pytest.fixture(autouse=True)
    def clean_up(self, project):
        yield
        with project.adapter.connection_named("__test"):
            relation = project.adapter.Relation.create(
                database=project.database, schema=f"{project.test_schema}_seeds"
            )
            project.adapter.drop_schema(relation)

            relation = project.adapter.Relation.create(
                database=project.database, schema=project.test_schema
            )
            project.adapter.drop_schema(relation)


class TestFabricCloneSameTargetAndState(BaseCloneSameTargetAndState):
    pass



from typing import Any, ClassVar

from dbt.adapters.base import Column


class FabricColumn(Column):
    TYPE_LABELS: ClassVar[dict[str, str]] = {
        "BINARY": "BINARY(1)",
        "BOOLEAN": "BIT",
        "CHAR": "CHAR(1)",
        "DATETIME2": "DATETIME2(6)",
        "INTEGER": "INT",
        "MONEY": "DECIMAL",
        "NCHAR": "CHAR(1)",
        "NVARCHAR": "VARCHAR(MAX)",
        "SMALLMONEY": "DECIMAL",
        "STRING": "VARCHAR(MAX)",
        "TIME": "TIME(6)",
        "TIMESTAMP": "DATETIME2(6)",
        "TINYINT": "SMALLINT",
        "VARBINARY": "VARBINARY(MAX)",
        "VARCHAR": "VARCHAR(MAX)",
    }

    @classmethod
    def string_type(cls, size: int) -> str:
        return f"varchar({size if size > 0 else 'max'})"

    def literal(self, value: Any) -> str:
        return "cast('{}' as {})".format(value, self.data_type)

    def is_integer(self):
        return super().is_integer() or self.dtype.lower() == "int"

    def is_string(self) -> bool:
        return self.dtype.lower() in ["varchar", "char"]

    def is_numeric(self) -> bool:
        return self.dtype.lower() in ["numeric", "decimal", "money", "smallmoney"]

    @property
    def data_type(self) -> str:
        if self.dtype == "datetime2":
            return f"datetime2({self.numeric_scale})"
        return super().data_type



import pytest

from dbt.tests.adapter.python_model.test_python_model import (
    BasePythonIncrementalTests,
    BasePythonModelTests,
)
from dbt.tests.adapter.python_model.test_spark import BasePySparkTests


class TestPythonModelTestsFabric(BasePythonModelTests):
    pass


@pytest.mark.skip("Support for incremental Python models is not yet implemented")
class TestPythonIncrementalTestsFabric(BasePythonIncrementalTests):
    pass


@pytest.mark.skip("We only support PySpark DataFrames for now")
class TestPySparkTestsFabric(BasePySparkTests):
    pass



import pytest

from dbt.tests.adapter.grants.test_incremental_grants import BaseIncrementalGrants
from dbt.tests.adapter.grants.test_invalid_grants import BaseInvalidGrants
from dbt.tests.adapter.grants.test_model_grants import BaseModelGrants
from dbt.tests.adapter.grants.test_seed_grants import BaseSeedGrants
from dbt.tests.adapter.grants.test_snapshot_grants import BaseSnapshotGrants


@pytest.mark.grants
class TestModelGrantsFabric(BaseModelGrants):
    pass


@pytest.mark.grants
class TestSeedGrantsFabric(BaseSeedGrants):
    pass


@pytest.mark.grants
class TestSnapshotGrantsFabric(BaseSnapshotGrants):
    pass


@pytest.mark.grants
class TestIncrementalGrantsFabric(BaseIncrementalGrants):
    pass


@pytest.mark.grants
class TestInvalidGrantsFabric(BaseInvalidGrants):
    def privilege_does_not_exist_error(self):
        return "Incorrect syntax near"

    def grantee_does_not_exist_error(self):
        return "could not be"



import pytest

from dbt.tests.adapter.utils.test_array_append import BaseArrayAppend
from dbt.tests.adapter.utils.test_array_concat import BaseArrayConcat
from dbt.tests.adapter.utils.test_array_construct import BaseArrayConstruct


@pytest.mark.skip(reason="Array concat is not supported in Fabric")
class TestArrayAppendFabric(BaseArrayAppend):
    pass


@pytest.mark.skip(reason="Array concat is not supported in Fabric")
class TestArrayConcatFabric(BaseArrayConcat):
    pass


class TestArrayConstructFabric(BaseArrayConstruct):
    pass



from dataclasses import dataclass

from dbt_common.dataclass_schema import StrEnum

from dbt.adapters.base.relation import Policy


class FabricRelationType(StrEnum):
    Table = "table"
    View = "view"
    Ephemeral = "ephemeral"
    CTE = "cte"


class FabricIncludePolicy(Policy):
    database: bool = True
    schema: bool = True
    identifier: bool = True


@dataclass
class FabricQuotePolicy(Policy):
    database: bool = True
    schema: bool = True
    identifier: bool = True



import time
import urllib.parse
from typing import Any, Self

import dbt_common.exceptions
import requests

from dbt.adapters.fabric.fabric_credentials import FabricCredentials
from dbt.adapters.fabric.fabric_token_provider import FabricTokenProvider


class FabricApiClient:
    _LIVY_API_VERSION = "2023-12-01"
    _WAREHOUSE_SNAPSHOT_TIMEOUT_SECONDS = 60 * 30  # 30 minutes
    _instance: Self | None = None

    def __init__(
        self, credentials: FabricCredentials, token_provider: FabricTokenProvider
    ) -> None:
        self._credentials = credentials
        self._token_provider = token_provider
        self._warehouse_connection_string: str | None = None
        self._lakehouse_id: str | None = None
        self._warehouse_id: str | None = None
        self._workspace_id: str | None = None
        self._cached_warehouses: list[dict] | None = None
        self._cached_lakehouses: list[dict] | None = None
        self._livy_session_id: str | None = None
        self._warehouse_snapshot_operations: dict[str, str] = {}

    @classmethod
    def create(cls, credentials: FabricCredentials, token_provider: FabricTokenProvider) -> Self:
        if cls._instance is None:
            cls._instance = FabricApiClient(credentials, token_provider)
        return cls._instance

    def _get_auth_headers(self) -> dict[str, str]:
        token = self._token_provider.get_api_token()
        return {
            "Authorization": f"Bearer {token}",
            "Accept": "application/json",
        }

    def _api_request(
        self, url: str, method: str = "get", body: dict | None = None
    ) -> requests.Response:
        response = requests.request(method, url, json=body, headers=self._get_auth_headers())

        if response.status_code == 429:
            retry_after = int(response.headers.get("Retry-After", 5))
            time.sleep(retry_after)
            return self._api_request(url, method, body)

        if not response.status_code >= 200 and response.status_code < 300:
            raise dbt_common.exceptions.DbtRuntimeError(
                f"{method} request to {url} failed with status code {response.status_code}: {response.text}"
            )
        return response

    def _api_get(self, url: str) -> requests.Response:
        return self._api_request(url, method="get")

    def _api_post(self, url: str, body: dict) -> requests.Response:
        return self._api_request(url, method="post", body=body)

    def _api_patch(self, url: str, body: dict) -> requests.Response:
        return self._api_request(url, method="patch", body=body)

    def _api_delete(self, url: str) -> requests.Response:
        return self._api_request(url, method="delete")

    def get_workspace_id(self) -> str:
        if self._workspace_id is not None:
            return self._workspace_id
        if self._credentials.workspace_id:
            return self._credentials.workspace_id
        if not self._credentials.workspace_name:
            raise dbt_common.exceptions.DbtConfigError(
                "Either workspace_id or workspace_name must be provided."
            )

        query_param = f"name eq '{self._credentials.workspace_name}'"
        query_param_encoded = urllib.parse.quote_plus(query_param)
        response = self._api_get(
            f"{self._credentials.powerbi_base_api_uri}/myorg/groups?$filter={query_param_encoded}"
        )
        workspaces = response.json().get("value", [])

        if len(workspaces) == 0:
            raise dbt_common.exceptions.DbtRuntimeError(
                f"No workspace found with name {self._credentials.workspace_name}"
            )

        self._workspace_id = workspaces[0]["id"]
        assert self._workspace_id is not None
        return self._workspace_id

    def get_warehouses(self, fetch_all: bool = True) -> list[dict]:
        if self._cached_warehouses is not None:
            return self._cached_warehouses

        workspace_id = self.get_workspace_id()

        url = f"{self._credentials.fabric_base_api_uri}/workspaces/{workspace_id}/warehouses"
        warehouses = []

        while url is not None:
            response = self._api_get(url)
            warehouses = warehouses + response.json().get("value", [])
            url = response.json().get("continuationUri", None) if fetch_all else None

        if fetch_all:
            self._cached_warehouses = warehouses
        return warehouses

    def get_lakehouses(self, fetch_all: bool = True) -> list[dict]:
        if self._cached_lakehouses is not None:
            return self._cached_lakehouses

        workspace_id = self.get_workspace_id()

        url = f"{self._credentials.fabric_base_api_uri}/workspaces/{workspace_id}/lakehouses"
        lakehouses = []

        while url is not None:
            response = self._api_get(url)
            lakehouses = lakehouses + response.json().get("value", [])
            url = response.json().get("continuationUri", None) if fetch_all else None

        if fetch_all:
            self._cached_lakehouses = lakehouses
        return lakehouses

    def get_warehouse_connection_string(self) -> str:
        if self._warehouse_connection_string is not None:
            return self._warehouse_connection_string

        # first we try to find it in any warehouse (they all have the same connection string)
        warehouses = self.get_warehouses(fetch_all=False)
        if len(warehouses) > 0:
            self._warehouse_connection_string = warehouses[0]["properties"]["connectionString"]
            assert self._warehouse_connection_string is not None
            return self._warehouse_connection_string

        # then we try to find it in any lakehouse (also have the same connection string)
        lakehouses = self.get_lakehouses(fetch_all=False)
        if len(lakehouses) > 0:
            self._warehouse_connection_string = lakehouses[0]["properties"][
                "sqlEndpointProperties"
            ]["connectionString"]
            assert self._warehouse_connection_string is not None
            return self._warehouse_connection_string

        raise dbt_common.exceptions.DbtRuntimeError(
            f"No Data Warehouses or Lakehouses found in workspace"
        )

    def get_lakehouse_id(self) -> str:
        if self._lakehouse_id is not None:
            return self._lakehouse_id
        if self._credentials.lakehouse_id:
            return self._credentials.lakehouse_id
        if not self._credentials.lakehouse_name:
            raise dbt_common.exceptions.DbtConfigError(
                "Either lakehouse_id or lakehouse_name must be provided."
            )

        for lakehouse in self.get_lakehouses(self._credentials):
            if lakehouse["displayName"] == self._credentials.lakehouse_name:
                self._lakehouse_id = lakehouse["id"]
                assert self._lakehouse_id is not None
                return self._lakehouse_id

        raise dbt_common.exceptions.DbtRuntimeError(
            f"No Lakehouse found with name {self._credentials.lakehouse_name}"
        )

    def get_warehouse_id(self) -> str:
        if self._warehouse_id is not None:
            return self._warehouse_id

        for warehouse in self.get_warehouses():
            if warehouse["displayName"] == self._credentials.database:
                self._warehouse_id = warehouse["id"]
                assert self._warehouse_id is not None
                return self._warehouse_id

        raise dbt_common.exceptions.DbtRuntimeError(
            f"No Data Warehouse found with name {self._credentials.database}"
        )

    def get_warehouse_snapshots(self) -> list[dict]:
        warehouse_id = self.get_warehouse_id()
        workspace_id = self.get_workspace_id()

        url = (
            f"{self._credentials.fabric_base_api_uri}/workspaces/{workspace_id}/warehousesnapshots"
        )
        snapshots = []

        while url is not None:
            response = self._api_get(url)
            for snapshot in response.json().get("value", []):
                parent_warehouse_id = snapshot.get("properties", {}).get("parentWarehouseId")
                if parent_warehouse_id == warehouse_id:
                    snapshots.append(snapshot)

            url = response.json().get("continuationUri", None)

        return snapshots

    def create_warehouse_snapshot(
        self, snapshot_name: str, description: str | None = None
    ) -> None:
        url = f"{self._credentials.fabric_base_api_uri}/workspaces/{self.get_workspace_id()}/warehousesnapshots"
        body = {
            "displayName": snapshot_name,
            "creationPayload": {"parentWarehouseId": self.get_warehouse_id()},
        }
        if description is not None:
            body["description"] = description

        response = self._api_post(
            url,
            body,
        )

        location_uri = response.headers.get("Location")
        if location_uri is not None and response.status_code == 202:
            self._warehouse_snapshot_operations[snapshot_name] = location_uri

    def update_warehouse_snapshot(
        self, snapshot_id: str, snapshot_name: str, description: str | None = None
    ) -> None:
        url = f"{self._credentials.fabric_base_api_uri}/workspaces/{self.get_workspace_id()}/warehousesnapshots/{snapshot_id}"
        body: dict[str, Any] = {"properties": {}}
        if description is not None:
            body["description"] = description
        response = self._api_patch(url, body)

        location_uri = response.headers.get("Location")
        if location_uri is not None and response.status_code == 202:
            self._warehouse_snapshot_operations[snapshot_name] = location_uri

    def wait_and_get_snapshot_id_from_operation(self, operation_uri: str) -> str:
        timer = time.time()
        while True:
            if time.time() - timer > self._WAREHOUSE_SNAPSHOT_TIMEOUT_SECONDS:
                raise dbt_common.exceptions.DbtRuntimeError(
                    f"Timed out waiting for Warehouse Snapshot operation to complete after {self._WAREHOUSE_SNAPSHOT_TIMEOUT_SECONDS} seconds."
                )

            response = self._api_get(operation_uri)
            operation_status = response.json().get("status", "Unknown")
            retry_sleep = int(response.headers.get("Retry-After", 5))

            if operation_status == "Succeeded":
                result_location = response.headers["Location"]
                result_response = self._api_get(result_location)
                return result_response.json()["id"]

            if operation_status not in ("NotStarted", "Running"):
                raise dbt_common.exceptions.DbtRuntimeError(
                    f"Warehouse Snapshot operation failed with status {operation_status}."
                )

            time.sleep(retry_sleep)

    def create_or_update_warehouse_snapshot(
        self, snapshot_name: str, description: str | None = None
    ) -> None:
        existing_snapshot_id = None

        snapshot_operation_uri = self._warehouse_snapshot_operations.get(snapshot_name)
        if snapshot_operation_uri is not None:
            existing_snapshot_id = self.wait_and_get_snapshot_id_from_operation(
                snapshot_operation_uri
            )
        else:
            all_snapshots = self.get_warehouse_snapshots()
            for snapshot in all_snapshots:
                if snapshot["displayName"] == snapshot_name:
                    existing_snapshot_id = snapshot["id"]
                    break

        if existing_snapshot_id is not None:
            self.update_warehouse_snapshot(existing_snapshot_id, snapshot_name, description)
        else:
            self.create_warehouse_snapshot(snapshot_name, description)

    def delete_warehouse_snapshot(self, snapshot_name: str) -> None:
        for snapshot in self.get_warehouse_snapshots():
            if snapshot["displayName"] == snapshot_name:
                self._api_delete(
                    f"{self._credentials.fabric_base_api_uri}/workspaces/{self.get_workspace_id()}/warehousesnapshots/{snapshot['id']}"
                )

    def get_livy_base_api_uri(self) -> str:
        workspace_id = self.get_workspace_id()
        lakehouse_id = self.get_lakehouse_id()
        return f"{self._credentials.fabric_base_api_uri}/workspaces/{workspace_id}/lakehouses/{lakehouse_id}/livyapi/versions/{self._LIVY_API_VERSION}"

    def get_existing_livy_session(self) -> str | None:
        url = self.get_livy_base_api_uri() + "/sessions"
        response = self._api_get(url)
        sessions = response.json().get("value", [])
        for session in sessions:
            if session["name"] == "dbt-fabric" and session["livyState"] in (
                "idle",
                "starting",
                "running",
            ):
                return session["id"]
        return None

    def initialize_livy_session(self) -> str:
        url = self.get_livy_base_api_uri() + "/sessions"
        response = self._api_post(url, {"name": "dbt-fabric"})
        return response.json()["id"]

    def get_livy_session_id(self) -> str:
        if self._livy_session_id is None:
            self._livy_session_id = (
                self.get_existing_livy_session() or self.initialize_livy_session()
            )
        return self._livy_session_id

    def get_livy_session_base_uri(self) -> str:
        return self.get_livy_base_api_uri() + f"/sessions/{self.get_livy_session_id()}"

    def get_livy_session_state(self) -> str:
        response = self._api_get(self.get_livy_session_base_uri())
        return response.json().get("state", "unknown")

    def get_livy_statement(self, statement_id: str) -> dict[str, Any]:
        url = self.get_livy_session_base_uri() + f"/statements/{statement_id}"
        response = self._api_get(url)
        return response.json()

    def submit_livy_statement(self, code: str) -> str:
        url = self.get_livy_session_base_uri() + "/statements"
        response = self._api_post(url, {"code": code, "kind": "pyspark"})
        return response.json()["id"]



import pytest

from dbt.tests.adapter.basic import expected_catalog
from dbt.tests.adapter.basic.test_adapter_methods import BaseAdapterMethod
from dbt.tests.adapter.basic.test_base import BaseSimpleMaterializations
from dbt.tests.adapter.basic.test_docs_generate import (
    BaseDocsGenerate,
    BaseDocsGenReferences,
    ref_models__docs_md,
    ref_models__ephemeral_copy_sql,
    ref_models__ephemeral_summary_sql,
    ref_models__schema_yml,
    ref_sources__schema_yml,
)
from dbt.tests.adapter.basic.test_empty import BaseEmpty
from dbt.tests.adapter.basic.test_ephemeral import BaseEphemeral
from dbt.tests.adapter.basic.test_generic_tests import BaseGenericTests
from dbt.tests.adapter.basic.test_incremental import (
    BaseIncremental,
    BaseIncrementalNotSchemaChange,
)
from dbt.tests.adapter.basic.test_singular_tests import BaseSingularTests
from dbt.tests.adapter.basic.test_singular_tests_ephemeral import BaseSingularTestsEphemeral
from dbt.tests.adapter.basic.test_snapshot_check_cols import BaseSnapshotCheckCols
from dbt.tests.adapter.basic.test_snapshot_timestamp import BaseSnapshotTimestamp
from dbt.tests.adapter.basic.test_table_materialization import BaseTableMaterialization
from dbt.tests.adapter.basic.test_validate_connection import BaseValidateConnection


class TestSimpleMaterializations(BaseSimpleMaterializations):
    pass


class TestSingularTestsFabric(BaseSingularTests):
    pass


class TestSingularTestsEphemeralFabric(BaseSingularTestsEphemeral):
    @pytest.fixture(scope="class")
    def project_config_update(self):
        return {
            "name": "singular_tests_ephemeral",
            "models": {
                "+materialized": "table"  # no support for nested CTEs in Views yet
            },
        }


class TestEmptyFabric(BaseEmpty):
    pass


class TestEphemeralFabric(BaseEphemeral):
    pass


class TestIncrementalFabric(BaseIncremental):
    pass


class TestIncrementalNotSchemaChangeFabric(BaseIncrementalNotSchemaChange):
    @pytest.fixture(scope="class")
    def models(self):
        return {
            "incremental_not_schema_change.sql": """
{{ config(materialized="incremental",
unique_key="user_id_current_time",on_schema_change="sync_all_columns") }}
select
CAST(1 + '-' + current_timestamp AS DATETIME2(6)) as user_id_current_time,
{% if is_incremental() %}
'thisis18characters' as platform
{% else %}
'okthisis20characters' as platform
{% endif %}
"""
        }


class TestGenericTestsFabric(BaseGenericTests):
    pass


class TestSnapshotCheckColsFabric(BaseSnapshotCheckCols):
    pass


class TestSnapshotTimestampFabric(BaseSnapshotTimestamp):
    pass


class TestBaseCachingFabric(BaseAdapterMethod):
    pass


class TestValidateConnectionFabric(BaseValidateConnection):
    pass


class TestDocsGenerateFabric(BaseDocsGenerate):
    @pytest.fixture(scope="class")
    def expected_catalog(self, project, profile_user):
        return expected_catalog.base_expected_catalog(
            project,
            role=profile_user,
            id_type="int",
            text_type="varchar",
            time_type="datetime2",
            view_type="VIEW",
            table_type="BASE TABLE",
            model_stats=expected_catalog.no_stats(),
        )


class TestDocsGenReferencesFabric(BaseDocsGenReferences):
    @pytest.fixture(scope="class")
    def expected_catalog(self, project, profile_user):
        return expected_catalog.expected_references_catalog(
            project,
            role=profile_user,
            id_type="int",
            text_type="varchar",
            time_type="datetime2",
            bigint_type="int",
            view_type="VIEW",
            table_type="BASE TABLE",
            model_stats=expected_catalog.no_stats(),
        )

    @pytest.fixture(scope="class")
    def models(self):
        ref_models__view_summary_sql = """
{{
  config(
    materialized = "view"
  )
}}

select first_name, ct from {{ref('ephemeral_summary')}}
"""

        return {
            "schema.yml": ref_models__schema_yml,
            "sources.yml": ref_sources__schema_yml,
            "view_summary.sql": ref_models__view_summary_sql,
            "ephemeral_summary.sql": ref_models__ephemeral_summary_sql,
            "ephemeral_copy.sql": ref_models__ephemeral_copy_sql,
            "docs.md": ref_models__docs_md,
        }


class TestTableMaterializationFabric(BaseTableMaterialization):
    pass



import os

import pytest

from dbt.adapters.fabric.fabric_adapter import FabricAdapter
from dbt.tests.util import run_dbt, run_sql_with_adapter


class TestSchemaCreation:
    @pytest.fixture(scope="class")
    def test_user(self):
        return os.getenv("DBT_TEST_USER_1", "dbo")

    @pytest.fixture(scope="class")
    def models(self):
        return {
            "dummy.sql": """
{{ config(schema='with_custom_auth') }}
select 1 as id
""",
        }

    @pytest.fixture(scope="class")
    def dbt_profile_target_update(self, test_user: str):
        return {"schema_authorization": f"{test_user}"}

    @pytest.fixture(scope="class")
    def initialization(initialization: None, adapter: FabricAdapter, test_user: str) -> None:
        if test_user != "dbo":
            # Fabric only seems to recognize users as soon as you have granted them something
            # You cannot grant something to dbo
            run_sql_with_adapter(adapter, f"grant select on schema :: dbo to [{test_user}]")

    @staticmethod
    def _verify_schema_owner(schema_name, owner, project):
        get_schema_owner = f"""
select p.name
from sys.schemas s
join sys.database_principals p
on s.principal_id = p.principal_id
where s.name = '{schema_name}'
        """
        result = project.run_sql(get_schema_owner, fetch="one")
        first_record = result[0]
        assert first_record == owner, f"Schema owner for {schema_name} is not {owner}"

    def test_schema_creation(self, project, unique_schema, test_user: str):
        res = run_dbt(["run"])
        assert len(res) == 1

        self._verify_schema_owner(unique_schema, test_user, project)
        self._verify_schema_owner(unique_schema + "_with_custom_auth", test_user, project)



import pytest

from tests.functional.packages.base_package_test import BaseDbtPackageTests


class TestDbtUtils(BaseDbtPackageTests):
    @pytest.fixture(scope="class")
    def macros(self):
        return {
            "limit_zero.sql": """
{% macro fabric__limit_zero() %}
  {{ return('where 0=1') }} 
{% endmacro %}"""
        }

    @pytest.fixture(scope="class")
    def package_name(self) -> str:
        return "dbt_utils"

    @pytest.fixture(scope="class")
    def package_repo(self) -> str:
        return "https://github.com/dbt-labs/dbt-utils"

    @pytest.fixture(scope="class")
    def package_revision(self) -> str:
        return "1.3.0"

    @pytest.fixture(scope="class")
    def models_config(self):
        return {"dbt_utils_integration_tests": {"sql": {"test_groupby": {"+enabled": False}}}}

    @pytest.fixture(scope="class")
    def seeds_config(self):
        return {
            "dbt_utils_integration_tests": {
                "sql": {
                    "data_get_single_value": {
                        "+column_types": {"date_value": "datetime2(6)", "int_value": "int"}
                    }
                },
                "schema_tests": {
                    "data_test_sequential_timestamps": {
                        "+column_types": {"my_timestamp": "datetime2(6)"}
                    }
                },
            }
        }



from dbt.tests.adapter.relations.test_changing_relation_type import BaseChangeRelationTypeValidator
from dbt.tests.adapter.relations.test_dropping_schema_named import BaseDropSchemaNamed


class TestChangeRelationTypesFabric(BaseChangeRelationTypeValidator):
    pass


class TestDropSchemaNamedFabric(BaseDropSchemaNamed):
    pass



import pytest

from dbt.tests.adapter.utils.test_validate_sql import BaseValidateSqlMethod


@pytest.mark.skip("Fabric does not support SHOWPLAN at the moment")
class TestValidateSqlMethodFabric(BaseValidateSqlMethod):
    @pytest.fixture(scope="class")
    def valid_sql(self) -> str:
        return "select 1 as id"



import time
from typing import Any

from dbt.adapters.base import PythonJobHelper
from dbt.adapters.fabric.fabric_api_client import FabricApiClient
from dbt.adapters.fabric.fabric_credentials import FabricCredentials
from dbt.adapters.fabric.fabric_token_provider import FabricTokenProvider


class LivySessionResult:
    def __init__(self, success: bool, error_message: str | None = None) -> None:
        self.error_message = error_message
        self.success = success


class LivySession:
    _POLLING_INTERVAL = 5  # seconds
    _MAX_POLLING_ATTEMPTS = 60  # 5 minutes

    def __init__(self, fabric_api_client: FabricApiClient) -> None:
        self._fabric_api_client = fabric_api_client

    def _get_logs_url(self) -> str:
        return f"https://app.fabric.microsoft.com/workloads/de-ds/sparkmonitor/{self._fabric_api_client.get_lakehouse_id()}/{self._fabric_api_client.get_livy_session_id()}"

    def _wait_for_session_ready(self) -> None:
        attempts = 0
        while self._fabric_api_client.get_livy_session_state() != "idle":
            if attempts >= self._MAX_POLLING_ATTEMPTS:
                raise TimeoutError("Livy session did not become idle in time.")
            attempts += 1
            time.sleep(self._POLLING_INTERVAL)

    def _wait_for_statement_ready(self, statement_id: str) -> dict[str, Any]:
        attempts = 0
        while True:
            time.sleep(self._POLLING_INTERVAL)
            statement_response = self._fabric_api_client.get_livy_statement(statement_id)
            statement_state = statement_response.get("state", "unknown")
            if statement_state in ("available", "error"):
                return statement_response
            if attempts >= self._MAX_POLLING_ATTEMPTS:
                raise TimeoutError("Livy statement did not become available in time.")
            attempts += 1

    def submit(self, python_code: str) -> Any:
        try:
            self._wait_for_session_ready()
            statement_id = self._fabric_api_client.submit_livy_statement(python_code)
            response = self._wait_for_statement_ready(statement_id)
            return LivySessionResult(
                success=response["state"] == "available"
                and response.get("output", {}).get("status") == "ok",
                error_message=response.get("output", {}).get("evalue"),
            )
        except Exception as e:
            print(f"Logs available at {self._get_logs_url()}")
            return LivySessionResult(success=False, error_message=str(e))


class FabricLivyHelper(PythonJobHelper):
    _livy_session: LivySession | None = None
    _sql_endpoint: str | None = None

    def __init__(self, parsed_model: dict, credential: FabricCredentials) -> None:
        fabric_api_client = FabricApiClient.create(credential, FabricTokenProvider(credential))

        if not self._livy_session:
            self._livy_session = LivySession(fabric_api_client)

        if not self._sql_endpoint:
            self._sql_endpoint = fabric_api_client.get_warehouse_connection_string()

    def submit(self, compiled_code: str) -> Any:
        assert self._livy_session is not None
        assert self._sql_endpoint is not None
        compiled_code = compiled_code.replace("DBT_FABRIC_REPLACED_WITH_HOST", self._sql_endpoint)
        return self._livy_session.submit(compiled_code)



import pytest

from dbt.tests.adapter.hooks.test_model_hooks import (
    BaseDuplicateHooksInConfigs,
    BaseHookRefs,
    BaseHooksRefsOnSeeds,
    BasePrePostModelHooksInConfig,
    BasePrePostModelHooksInConfigKwargs,
    BasePrePostModelHooksOnSeeds,
    BasePrePostModelHooksOnSeedsPlusPrefixed,
    BasePrePostModelHooksOnSeedsPlusPrefixedWhitespace,
    BasePrePostModelHooksOnSnapshots,
    BasePrePostSnapshotHooksInConfigKwargs,
)
from dbt.tests.adapter.hooks.test_run_hooks import (
    BaseAfterRunHooks,
)
from dbt.tests.fixtures.project import TestProjInfo


class RunModelFile:
    @pytest.fixture(scope="class", autouse=True)
    def setUp(self, project: TestProjInfo):
        project.run_sql("drop table if exists {schema}.on_model_hook;")
        project.run_sql("""
create table {schema}.on_model_hook (
    test_state       varchar(100), -- start|end
    target_dbname    varchar(100),
    target_host      varchar(100),
    target_name      varchar(100),
    target_schema    varchar(100),
    target_type      varchar(100),
    target_user      varchar(100),
    target_pass      varchar(100),
    target_threads   int,
    run_started_at   varchar(100),
    invocation_id    varchar(100),
    thread_id        varchar(100)
);
""")


class FabricHooksChecks:
    def check_hooks(self, state, project, host, count=1):
        ctxs = self.get_ctx_vars(state, count=count, project=project)
        for ctx in ctxs:
            assert ctx["test_state"] == state
            assert ctx["target_name"] == "default"
            assert ctx["target_schema"] == project.test_schema
            assert ctx["target_type"] == "fabric"

            assert ctx["run_started_at"] is not None and len(ctx["run_started_at"]) > 0, (
                "run_started_at was not set"
            )
            assert ctx["invocation_id"] is not None and len(ctx["invocation_id"]) > 0, (
                "invocation_id was not set"
            )
            assert ctx["thread_id"].startswith("Thread-")


class TestDuplicateHooksInConfigsFabric(BaseDuplicateHooksInConfigs):
    pass


class TestHookRefsFabric(RunModelFile, FabricHooksChecks, BaseHookRefs):
    pass


class TestHooksRefsOnSeedsFabric(BaseHooksRefsOnSeeds):
    pass


class TestPrePostModelHooksInConfigFabric(
    RunModelFile, FabricHooksChecks, BasePrePostModelHooksInConfig
):
    pass


class TestPrePostModelHooksInConfigKwargsFabric(
    RunModelFile, FabricHooksChecks, BasePrePostModelHooksInConfigKwargs
):
    pass


class TestPrePostModelHooksOnSeedsFabric(BasePrePostModelHooksOnSeeds):
    @pytest.fixture(scope="class")
    def project_config_update(self):
        return {
            "seed-paths": ["seeds"],
            "models": {},
            "seeds": {
                "post-hook": [
                    "alter table {{ this }} add new_col int",
                    "update {{ this }} set new_col = 1",
                    # call any macro to track dependency: https://github.com/dbt-labs/dbt-core/issues/6806
                    "select cast(null as {{ dbt.type_int() }}) as id",
                ],
                "quote_columns": False,
            },
        }


class TestPrePostModelHooksOnSeedsPlusPrefixedFabric(BasePrePostModelHooksOnSeedsPlusPrefixed):
    @pytest.fixture(scope="class")
    def project_config_update(self):
        return {
            "seed-paths": ["seeds"],
            "models": {},
            "seeds": {
                "+post-hook": [
                    "alter table {{ this }} add new_col int",
                    "update {{ this }} set new_col = 1",
                ],
                "quote_columns": False,
            },
        }


class TestPrePostModelHooksOnSeedsPlusPrefixedWhitespaceFabric(
    BasePrePostModelHooksOnSeedsPlusPrefixedWhitespace
):
    @pytest.fixture(scope="class")
    def project_config_update(self):
        return {
            "seed-paths": ["seeds"],
            "models": {},
            "seeds": {
                "+post-hook": [
                    "alter table {{ this }} add new_col int",
                    "update {{ this }} set new_col = 1",
                ],
                "quote_columns": False,
            },
        }


class TestPrePostModelHooksOnSnapshotsFabric(BasePrePostModelHooksOnSnapshots):
    @pytest.fixture(scope="class")
    def project_config_update(self):
        return {
            "seed-paths": ["seeds"],
            "snapshot-paths": ["test-snapshots"],
            "models": {},
            "snapshots": {
                "post-hook": [
                    "alter table {{ this }} add new_col int",
                    "update {{ this }} set new_col = 1",
                ]
            },
            "seeds": {
                "quote_columns": False,
            },
        }


class TestPrePostSnapshotHooksInConfigKwargsFabric(BasePrePostSnapshotHooksInConfigKwargs):
    @pytest.fixture(scope="class")
    def project_config_update(self):
        return {
            "seed-paths": ["seeds"],
            "snapshot-paths": ["test-kwargs-snapshots"],
            "models": {},
            "snapshots": {
                "post-hook": [
                    "alter table {{ this }} add new_col int",
                    "update {{ this }} set new_col = 1",
                ]
            },
            "seeds": {
                "quote_columns": False,
            },
        }


class TestAfterRunHooksFabric(BaseAfterRunHooks):
    pass



import pytest

from dbt.tests.adapter.utils.test_source_freshness_custom_info import BaseCalculateFreshnessMethod


class TestCalculateFreshnessMethodFabric(BaseCalculateFreshnessMethod):
    @pytest.fixture(scope="class")
    def valid_sql(self) -> str:
        return "select current_timestamp as c"



from typing import Iterable

from dbt.tests.adapter.simple_snapshot.test_snapshot import (
    BaseSimpleSnapshot,
    BaseSimpleSnapshotBase,
    BaseSnapshotCheck,
)
from dbt.tests.util import relation_from_name, run_dbt


class BaseSimpleSnapshotBaseFabric(BaseSimpleSnapshotBase):
    def _assert_results(
        self,
        ids_with_current_snapshot_records: Iterable,
        ids_with_closed_out_snapshot_records: Iterable,
    ):
        records = set(
            self.get_snapshot_records(
                "id, case when dbt_valid_to is null then 1 else 0 end as is_current"
            )
        )
        expected_records = set().union(
            {(i, 1) for i in ids_with_current_snapshot_records},
            {(i, 0) for i in ids_with_closed_out_snapshot_records},
        )
        for record in records:
            assert record in expected_records

    def add_fact_column(self, column: str = None, definition: str = None):  # type: ignore
        table_name = relation_from_name(self.project.adapter, "fact")
        sql = f"""
            alter table {table_name}
            add {column} {definition}
        """
        self.project.run_sql(sql)


class TestSimpleSnapshotFabric(BaseSimpleSnapshotBaseFabric, BaseSimpleSnapshot):
    def test_new_column_captured_by_snapshot(self, project):
        self.add_fact_column("full_name", "varchar(200)")
        self.update_fact_records(
            {
                "full_name": "first_name + ' ' + last_name",
                "updated_at": "dateadd(day, 1, updated_at)",
            },
            "id between 11 and 20",
        )
        run_dbt(["snapshot"])
        self._assert_results(
            ids_with_current_snapshot_records=range(1, 21),
            ids_with_closed_out_snapshot_records=range(11, 21),
        )

    def test_updates_are_captured_by_snapshot(self, project):
        self.update_fact_records(
            {"updated_at": "dateadd(day, 1, updated_at)"}, "id between 16 and 20"
        )
        run_dbt(["snapshot"])
        self._assert_results(
            ids_with_current_snapshot_records=range(1, 21),
            ids_with_closed_out_snapshot_records=range(16, 21),
        )


class TestSnapshotCheckFabric(BaseSimpleSnapshotBaseFabric, BaseSnapshotCheck):
    pass



import pytest

from dbt.tests.adapter.simple_copy.test_copy_uppercase import BaseSimpleCopyUppercase
from dbt.tests.adapter.simple_copy.test_simple_copy import EmptyModelsArentRunBase, SimpleCopyBase


class TestEmptyModelsArentRunFabric(EmptyModelsArentRunBase):
    pass


class TestSimpleCopyBaseFabric(SimpleCopyBase):
    def test_simple_copy_with_materialized_views(self, project):
        pytest.skip("Materialized views are not supported in this adapter.")


class TestSimpleCopyUppercaseFabric(BaseSimpleCopyUppercase):
    @pytest.fixture(scope="class")
    def dbt_profile_target(self, dbt_profile_target):
        return dbt_profile_target



import pytest

from dbt.adapters.fabric.fabric_api_client import FabricApiClient
from dbt.tests.util import run_dbt


class TestWarehouseSnapshots:
    @pytest.fixture(scope="class")
    def snapshot_name(self, unique_schema: str):
        return f"{unique_schema}_snapshot"

    @pytest.fixture(scope="class")
    def models(self):
        return {"simple.sql": "select 1 as id"}

    @pytest.fixture(scope="class")
    def snapshot_description(self) -> str:
        return "A snapshot of the warehouse state during an automated test run."

    @pytest.fixture(scope="class")
    def project_config_update(self, snapshot_name: str, snapshot_description: str):
        return {
            "on-run-start": [
                f"{{{{ create_or_update_fabric_warehouse_snapshot('{snapshot_name}', '{snapshot_description}') }}}}"
            ],
            "on-run-end": [
                f"{{{{ create_or_update_fabric_warehouse_snapshot('{snapshot_name}', '{snapshot_description}') }}}}"
            ],
        }

    @pytest.fixture(autouse=True)
    def clean_up(self, fabric_api_client: FabricApiClient, snapshot_name: str):
        fabric_api_client.delete_warehouse_snapshot(snapshot_name)
        yield
        fabric_api_client.delete_warehouse_snapshot(snapshot_name)

    def test_warehouse_snapshots(
        self,
        project,
        fabric_api_client: FabricApiClient,
        snapshot_name: str,
        snapshot_description: str,
    ):
        results = run_dbt(["run"])
        assert len(results) == 3
        all_snapshots = fabric_api_client.get_warehouse_snapshots()
        assert any(
            snapshot["displayName"] == snapshot_name
            and snapshot["description"] == snapshot_description
            for snapshot in all_snapshots
        )



from dataclasses import dataclass
from typing import Any

import agate

from dbt.adapters.base.relation import Policy
from dbt.adapters.contracts.relation import RelationConfig
from dbt.adapters.fabric.relation_configs.policies import FabricIncludePolicy, FabricQuotePolicy
from dbt.adapters.relation_configs import RelationConfigBase, RelationResults


@dataclass(frozen=True, eq=True, unsafe_hash=True)
class FabricRelationConfigBase(RelationConfigBase):
    """
    This base class implements a few boilerplate methods and provides some light structure for Fabric relations.
    """

    @classmethod
    def include_policy(cls) -> Policy:
        return FabricIncludePolicy()

    @classmethod
    def quote_policy(cls) -> Policy:
        return FabricQuotePolicy()

    @classmethod
    def from_relation_config(cls, relation_config: RelationConfig):
        relation_config_dict = cls.parse_relation_config(relation_config)
        relation = cls.from_dict(relation_config_dict)
        return relation

    @classmethod
    def parse_relation_config(cls, relation_config: RelationConfig) -> dict:
        raise NotImplementedError(
            "`parse_relation_config()` needs to be implemented on this RelationConfigBase instance"
        )

    @classmethod
    def from_relation_results(cls, relation_results: RelationResults):
        relation_config = cls.parse_relation_results(relation_results)
        relation = cls.from_dict(relation_config)
        return relation  # type: ignore

    @classmethod
    def parse_relation_results(cls, relation_results: RelationResults) -> dict[str, Any]:
        raise NotImplementedError(
            "`parse_relation_results()` needs to be implemented on this RelationConfigBase instance"
        )

    @classmethod
    def _get_first_row(cls, results: agate.Table) -> agate.Row:
        try:
            return results.rows[0]
        except IndexError:
            return agate.Row(values=set())



import pytest

from dbt.tests.adapter.utils.data_types.test_type_bigint import BaseTypeBigInt
from dbt.tests.adapter.utils.data_types.test_type_boolean import BaseTypeBoolean
from dbt.tests.adapter.utils.data_types.test_type_float import BaseTypeFloat
from dbt.tests.adapter.utils.data_types.test_type_int import BaseTypeInt
from dbt.tests.adapter.utils.data_types.test_type_numeric import BaseTypeNumeric
from dbt.tests.adapter.utils.data_types.test_type_string import BaseTypeString
from dbt.tests.adapter.utils.data_types.test_type_timestamp import (
    BaseTypeTimestamp,
    seeds__expected_csv,
)


@pytest.mark.skip(reason="SQL Server shows 'numeric' if you don't explicitly cast it to bigint")
class TestTypeBigIntFabric(BaseTypeBigInt):
    pass


class TestTypeFloatFabric(BaseTypeFloat):
    pass


class TestTypeIntFabric(BaseTypeInt):
    pass


class TestTypeNumericFabric(BaseTypeNumeric):
    pass


# Tests failing different than expected result
class TestTypeStringFabric(BaseTypeString):
    def assert_columns_equal(self, project, expected_cols, actual_cols):
        #  ignore the size of the varchar since we do
        #  an optimization to not use varchar(max) all the time
        assert expected_cols[:-1] == actual_cols[:-1], (
            f"Type difference detected: {expected_cols} vs. {actual_cols}"
        )


class TestTypeTimestampFabric(BaseTypeTimestamp):
    @pytest.fixture(scope="class")
    def seeds(self):
        seeds__expected_yml = """
version: 2
seeds:
  - name: expected
    config:
      column_types:
        timestamp_col: "datetime2(6)"
        """

        return {
            "expected.csv": seeds__expected_csv,
            "expected.yml": seeds__expected_yml,
        }


class TestTypeBooleanFabric(BaseTypeBoolean):
    pass



from dbt.tests.adapter.utils.test_current_timestamp import (
    BaseCurrentTimestamp,
    BaseCurrentTimestampNaive,
)


class TestCurrentTimestampFabric(BaseCurrentTimestamp):
    pass


class TestCurrentTimestampNaiveFabric(BaseCurrentTimestampNaive):
    pass
